import typing as t


class ModeGuards:
    run = lambda ctx: ctx.config.mode == "run"
    amend = lambda ctx: ctx.config.mode == "amend"


class HistoryGuards:
    failed = lambda ctx: ctx.history.get_status(ctx.project.project_id).status == "fail"
    skipped = lambda ctx: ctx.history.get_status(ctx.project.project_id).status == "skipped"
    succeeded = lambda ctx: ctx.history.get_status(ctx.project.project_id).status == "success"
    new = lambda ctx: ctx.history.get_status(ctx.project.project_id).status == "new"


class PreviousTaskGuards:
    is_success = lambda ctx: ctx.previous_return is None or ctx.previous_return.status == "processed"
    was_skipped = lambda ctx: ctx.previous_return is None or not ctx.previous_return.status == "skipped"


GuardsCollection: t.TypeAlias = ModeGuards | HistoryGuards | PreviousTaskGuards

MODE = ModeGuards
HISTORY = HistoryGuards
PREVIOUS_TASK = PreviousTaskGuards



import os
import typing as t
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from automyte.utils import bash

_DirName: t.TypeAlias = str
_FileName: t.TypeAlias = str
_FileContents: t.TypeAlias = str
_ProjectStructure: t.TypeAlias = dict[_DirName | _FileName, t.Union[_FileContents, "_ProjectStructure"]]


@pytest.fixture
def tmp_local_project_factory():
    """Setup a tmp project structure with folders, files and their contents for testing.

    Returns a factory function which accepts dictionary structure where we specify project structure.
    All projects are created in tmp folders which are cleaned up automatically after each run.

    Don't pass "dir" argument, as it is only used for recursive call for internal implementation.

    Example:
        tmp_local_project_factory(structure={
            'src': {
                'subdir1': {
                    'hello.txt': 'this will be the text for src/subdir1/hello.txt file',
                    'bye.py': 'print("good bye")',
                },
                'subdir2': {...},
            ...
            }
        })
    """
    rootdirs = []
    try:

        def _create_tmp_project(structure: _ProjectStructure, dir: str | None = None):
            if dir:  # Recursive call, just need to create child dirs.
                if not Path(dir).exists():
                    os.mkdir(dir)
                current_dir = dir
            else:  # First call, need to create TMP dir as parent and add it to array for removal.
                new_tmp_dir = TemporaryDirectory()
                rootdirs.append(new_tmp_dir)
                current_dir = new_tmp_dir.name

            for name, content in structure.items():
                if isinstance(content, str):  # Encountered a file.
                    with open(Path(current_dir) / name, "w") as f:
                        f.write(content)
                else:  # Encoruntered a folder, so need to generate the whole structure again, recursively.
                    _create_tmp_project(structure=content, dir=f"{current_dir}/{name}")

            return current_dir

        yield _create_tmp_project

    finally:
        for dir in rootdirs:
            dir.cleanup()


@pytest.fixture
def tmp_git_repo(tmp_local_project_factory):
    def _tmp_git_repo_factory(initial_structure, unstaged_structure=None):
        dir = tmp_local_project_factory(initial_structure)
        bash.execute(["git", "-C", dir, "init"])
        bash.execute(["git", "-C", dir, "add", "--all"])
        bash.execute(["git", "-C", dir, "commit", "-m", "Initial commit"])

        if unstaged_structure:
            tmp_local_project_factory(unstaged_structure, dir=dir)

        return dir

    return _tmp_git_repo_factory



from automyte.automaton import RunContext
from automyte.discovery import File


# TODO: Think of a better implementation?
class Breakpoint:
    def __call__(self, ctx: RunContext, file: File | None):
        import pdb

        pdb.set_trace()



import typing as t

from automyte.discovery import File
from automyte.history import AutomatonRunResult
from automyte.project import Project

from .run_context import RunContext
from .types import BaseTask, FileTask, InstructionForAutomaton, TaskReturn


class TasksFlow:
    def __init__(
        self, *args: list[FileTask], preprocess: list[BaseTask] | None = None, postprocess: list[BaseTask] | None = None
    ):
        self.preprocess_tasks = preprocess or []
        self.postprocess_tasks = postprocess or []
        self.tasks = list(*args)

    def execute(self, project: Project, ctx: "RunContext"):
        for preprocess_task in self.preprocess_tasks:
            instruction = self._handle_task_call(ctx=ctx, task=preprocess_task, file=None)

            if instruction == "skip":
                return AutomatonRunResult(status="skipped")
            elif instruction == "abort":
                return AutomatonRunResult(status="fail", error=str(ctx.previous_return.value))

        for file in project.explorer.explore():
            for process_file_task in self.tasks:
                instruction = self._handle_task_call(ctx=ctx, task=process_file_task, file=file)

                if instruction == "skip":
                    return AutomatonRunResult(status="skipped")
                elif instruction == "abort":
                    return AutomatonRunResult(status="fail", error=str(ctx.previous_return.value))

            ctx.cleanup_file_returns()

        # Has to be called prior to postprocess tasks, otherwise files changes are not reflected on disk before vcs calls.
        project.apply_changes()

        for post_task in self.postprocess_tasks:
            instruction = self._handle_task_call(ctx=ctx, task=post_task, file=None)

            if instruction == "skip":
                return AutomatonRunResult(status="skipped")
            elif instruction == "abort":
                return AutomatonRunResult(status="fail", error=str(ctx.previous_return.value))

        return AutomatonRunResult(status="success")

    def _handle_task_call(self, ctx: "RunContext", task: BaseTask, file: File | None) -> "InstructionForAutomaton":
        """Convenience wrapper for calling and handling all tasks.

        Wraps plain python values into TaskReturns,
            so that user doesn't have to do it unless they want to specify behaviour;
        Save task return into ctx;
        Return instruction for automaton on what to do next (like skip the project, continue or abort right away).

        If the task raised an Exception - save it's value into task return with "errored" status and instruct to abort.
        """
        try:
            task_result = wrap_task_result(task(ctx, file))

        except Exception as e:
            ctx.save_task_result(result=TaskReturn(instruction="abort", value=str(e), status="errored"), file=file)
            return "abort"

        else:
            ctx.save_task_result(result=task_result, file=file)
            return task_result.instruction


def wrap_task_result(value: t.Any) -> TaskReturn:
    if isinstance(value, TaskReturn):
        return value
    else:
        return TaskReturn(instruction="continue", status="processed", value=value)



import typing as t

from automyte import Automaton, Config, ContainsFilter, File, LocalFilesExplorer, Project, RunContext, TasksFlow, guards


def lol(ctx: RunContext, file: File):
    import re

    file.edit(re.sub(r"world", "there", file.get_contents()))


def test_guards_simple(tmp_local_project_factory):
    dir = tmp_local_project_factory(structure={"src": {"hello.txt": "hello world!"}})

    Automaton(
        name="impl1",
        config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
        projects=[
            Project(
                project_id="test_project",
                explorer=LocalFilesExplorer(rootdir=dir, filter_by=ContainsFilter(contains="hello world")),
            ),
        ],
        flow=TasksFlow(
            [
                guards.Conditional(
                    lol,
                    lol,
                    lol,
                    on=guards.MODE.amend,
                ),
            ]
        ),
    ).run()

    with open(f"{dir}/src/hello.txt", "r") as f:
        assert f.read() == "hello world!"



from pathlib import Path

import pytest

from automyte.automaton.run_context import RunContext
from automyte.config.base import Config
from automyte.history.types import AutomatonRunResult
from automyte.project import Project
from automyte.utils.random import random_hash
from automyte.vcs.base import VCS


@pytest.fixture
def run_ctx():
    def _ctx_factory(
        dir: str | Path, vcs: VCS | None = None, project: Project | None = None, automaton_name: str = "auto"
    ):
        if project is None:
            project = Project(project_id=random_hash(), rootdir=str(dir))
        if vcs is None:
            vcs = project.vcs

        return RunContext(
            automaton_name=automaton_name,
            config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
            vcs=vcs,
            project=project,
            current_status=AutomatonRunResult("running"),
            previous_status=AutomatonRunResult("new"),
            global_tasks_returns=[],
            file_tasks_returns=[],
        )

    return _ctx_factory



import logging.config

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {"detailed": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"}},
    "handlers": {"console": {"class": "logging.StreamHandler", "level": "DEBUG", "formatter": "detailed"}},
    "loggers": {
        "automyte": {  # Configure only your library's logger
            "level": "INFO",
            "handlers": ["console"],
            "propagate": False,
        }
    },
}



import typing as t

# from automyte.tasks.types import FileTask
from automyte.automaton import FileTask, RunContext
from automyte.discovery import File


class TaskGuard:
    def __init__(self, *args: FileTask) -> None:
        self.tasks = list(args)

    def __call__(self, ctx: RunContext, file: File):
        if self.guard(ctx):
            for task in self.tasks:
                result = task(ctx, file)
            else:
                # TODO: Need to actually properly handle cases when tasks fail.
                return locals().get("result", None)  # Just preventing linter from complaining
        else:
            # TODO: What should be done here? Supposedly, RunResult('skip')
            ...

    def guard(self, ctx: RunContext) -> bool:
        raise NotImplementedError


class Conditional(TaskGuard):
    def __init__(self, *args: FileTask, on: t.Callable[[RunContext], bool]) -> None:
        super().__init__(*args)
        self.validator = on

    def guard(self, ctx: RunContext) -> bool:
        return self.validator(ctx)



import typing as t
from pathlib import Path

import pytest

from automyte import OSFile
from automyte.utils.random import random_hash


@pytest.fixture
def tmp_os_file(tmp_local_project_factory):
    def _tmp_file_factory(contents: str, filename: str | None = None) -> OSFile:
        filename = filename or random_hash()
        dir = tmp_local_project_factory(structure={filename: contents})

        filepath = Path(dir) / filename
        with open(filepath, "w") as f:
            f.write(contents)

        return OSFile(fullname=str(filepath))

    return _tmp_file_factory
